from pyspark.sql import functions as F, DataFrame, Window
from tidal_per_transformers.transformers.utils import constants as c
from tidal_per_transformers.transformers import LoggableTransformer


class GenericRecommendationFormatterTransformer(LoggableTransformer):
    """
    Transformer to add standard format to a set of recommendations (e.g. userId, [recommendedIds])
    with a template that will specify the content type, the module and its version.
    """

    def __init__(self,
                 group_key: str,
                 recommendations_col: str,
                 item_type: str,
                 module_name: str,
                 module_version: str):
        super(GenericRecommendationFormatterTransformer, self).__init__()
        self.group_key = group_key
        self.recommendations_col = recommendations_col
        self.item_type = item_type
        self.module_name = module_name
        self.module_version = module_version

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return (dataset
                .withColumn(c.TYPE, F.lit(self.item_type))
                .withColumn(c.MODULE_ID, F.lit(f"{self.module_name}_v{self.module_version}"))
                .select(F.col(self.group_key).alias(c.HASH_KEY), c.TYPE, c.MODULE_ID,
                        F.posexplode(self.recommendations_col))
                .withColumn(self.recommendations_col,
                            F.collect_list(F.struct(F.col(c.TYPE), F.col("col").alias(c.ID))).over(
                                Window.partitionBy(c.HASH_KEY).orderBy("pos")))
                .groupby([c.HASH_KEY, c.MODULE_ID])
                .agg(F.max(self.recommendations_col).alias(self.recommendations_col))
                )



from pyspark.sql import functions as F, DataFrame
from pyspark.sql.window import Window
from tidal_per_transformers.transformers import LoggableTransformer


class CollectSessionsTransformer(LoggableTransformer):
    """
    Given a group of scored and time-sorted sessions by the users, it will collect
    them into 2 lists: recent sessions and relevant sessions. Both lists will be
    trimmed to the given limit.

    :param group_by: Columns to group the DataFrame by
    :param order_by_session: The column used to sort the values by time
    :param order_by_score: The column used to sort the values by score
    :param recent_alias: The name given to the new list column ordered by recency
    :param relevant_alias: The name given to the new list column ordered by relevance
    :param list_col: Column for which we collect the values
    :limit: Maximum number of items in the list
    """

    def __init__(self,
                 group_by: str,
                 order_by_session: str,
                 order_by_score: str,
                 recent_alias: str,
                 relevant_alias: str,
                 list_col: str,
                 limit: int = 300):
        super(CollectSessionsTransformer, self).__init__()
        self.group_by = group_by
        self.order_by_session = order_by_session
        self.order_by_score = order_by_score
        self.recent_alias = recent_alias
        self.relevant_alias = relevant_alias
        self.list_col = list_col
        self.limit = limit

    def _group_collect_recency_ordered(self, dataset: DataFrame) -> DataFrame:
        """
        Recent streaming sessions are composed of last X streamed tracks, with last
        listened item in last position.

        This gets built by going through user's sessions on an ascending fashion (most recent
        processed last) and incrementally collecting items, to then fetch the resulting list
        from the last instance.

        Note: due to the usage of F.last(), this function is not deterministic
        """
        window_session = Window.partitionBy(self.group_by).orderBy(F.col(self.order_by_session))
        return dataset.withColumn(
            self.recent_alias,
            F.reverse(
                F.slice(
                    F.reverse(
                        F.flatten(
                            F.collect_list(self.list_col).over(window_session)
                        )
                    ),
                    1, self.limit,
                )
            ),
        ).groupBy(F.col(self.group_by)).agg(
            F.last(F.col(self.recent_alias)).alias(self.recent_alias)
        )

    def _group_collect_relevance_ordered(self, dataset: DataFrame) -> DataFrame:
        """
        Relevant streaming sessions are composed by grouping of all sessions ordered by
        their relevance score, with the top session last and its first song in last position.

        This gets built by going through scored user's sessions descending in relevance (lowest
        scored processed last) and incrementally collecting items, to then fetch the resulting
        list from the last instance.

        Note: due to the usage of F.last(), this function is not deterministic
        """
        window_score = Window.partitionBy(self.group_by).orderBy(F.desc(F.col(self.order_by_score)))
        return dataset.withColumn(
            self.relevant_alias,
            F.reverse(
                F.slice(
                    F.flatten(F.collect_list(self.list_col).over(window_score)),
                    1, self.limit,
                )
            )
        ).groupBy(F.col(self.group_by)).agg(
            F.last(F.col(self.relevant_alias)).alias(self.relevant_alias)
        )

    def _transform(self, dataset: DataFrame) -> DataFrame:
        recent = self._group_collect_recency_ordered(dataset)
        relevant = self._group_collect_relevance_ordered(dataset)
        return recent.join(relevant, on=self.group_by)
